/**
 * Copyright (c) 2013 David I. Bell
 * Permission is granted to use, distribute, or modify this source,
 * provided that this copyright notice remains intact.
 */
package weather;

import java.util.GregorianCalendar;
import java.util.HashMap;
import java.util.TimeZone;


/**
 * A date-time class which allows the components to be easily extracted.
 * The date is always calculated as for GMT time but this isn't visible externally
 * and so can be used for other time zones if the date value is adjusted for them.
 * Fractional seconds are ignored.  This class is immutable.
 */
public class DateTime implements Comparable<DateTime>
{
	//
	// The individual parts of the date.
	// The months start with 1.
	//
	private int mYear;
	private int mMonth;
	private int mDay;
	private int mHour;
	private int mMinute;
	private int mSecond;

	/**
	 * The Date value corresponding to the individual parts above.
	 */
	private long mDateValue;

	/**
	 * The day of the week.
	 * No calculations are done with this value; it is only used for display.
	 */
	private int mDayOfWeek;

	/**
	 * The GMT time zone.
	 * This is used for consistency.
	 */
	private static final TimeZone mGMTimeZone = TimeZone.getTimeZone("GMT");

	/**
	 * The calendar instance that we reuse.
	 */
	private static final GregorianCalendar mCalendar = new GregorianCalendar();

	/**
	 * A hash map which maps a text representation of a date/time
	 * into a DateTime object.  This is used for date time values
	 * which are expected to be repeated.
	 */
	private static final HashMap<String, DateTime> mCachedDateTime =
		new HashMap<String, DateTime>(1000);

	/**
	 * Get a cached DateTime object for the specified text value.
	 * @param textValue
	 * @return
	 */
	public static DateTime getCachedDateTime(String textValue)
	{
		DateTime dateTime = mCachedDateTime.get(textValue);

		if (dateTime != null)
			return dateTime;

		textValue = Utils.getCachedString(textValue);

		dateTime = new DateTime(textValue);

		mCachedDateTime.put(textValue, dateTime);

		return dateTime;
	}


	/**
	 * Constructor which uses a GMT date/time value (i.e., from Date()).
	 * The fractional seconds are removed.
	 * @param dateValue
	 */
	public DateTime(long dateValue)
	{
		GregorianCalendar calendar = getCalendar();

		calendar.setTimeInMillis(dateValue);
		calendar.set(GregorianCalendar.MILLISECOND, 0);

		mDateValue = calendar.getTimeInMillis();

		mYear = calendar.get(GregorianCalendar.YEAR);
		mMonth = calendar.get(GregorianCalendar.MONTH) + 1;
		mDay = calendar.get(GregorianCalendar.DAY_OF_MONTH);
		mHour = calendar.get(GregorianCalendar.HOUR_OF_DAY);
		mMinute = calendar.get(GregorianCalendar.MINUTE);
		mSecond = calendar.get(GregorianCalendar.SECOND);
		mDayOfWeek = calendar.get(GregorianCalendar.DAY_OF_WEEK);
	}


	/**
	 * Constructor which uses a string formatted in one of the formats:
	 *	yyyymmddhhmm
	 *	yyyymmddhhmmss
	 * The numeric values are not checked for validity.
	 */
	public DateTime(String textValue)
	{
		if ((textValue == null) ||
			((textValue.length() != 12) && (textValue.length() != 14)))
		{
			throw new IllegalArgumentException("Bad date text value");
		}

		int year = Integer.parseInt(textValue.substring(0, 4));
		int month = Integer.parseInt(textValue.substring(4, 6));
		int day = Integer.parseInt(textValue.substring(6, 8));
		int hour = Integer.parseInt(textValue.substring(8, 10));
		int minute = Integer.parseInt(textValue.substring(10, 12));
		int second = 0;

		if (textValue.length() == 14)
			second = Integer.parseInt(textValue.substring(12, 14));

		setValues(year, month, day, hour, minute, second);
	}


	/**
	 * Constructor which uses broken-down date values.
	 * The months start with 1.
	 */
	public DateTime(int year, int month, int day, int hour, int minute, int second)
	{
		setValues(year, month, day, hour, minute, second);
	}


	/**
	 * Utility method to set the date values based on their components.
	 * @param year
	 * @param month
	 * @param day
	 * @param hour
	 * @param minute
	 * @param second
	 */
	private void setValues(int year, int month, int day, int hour, int minute, int second)
	{
		GregorianCalendar calendar = getCalendar();

		calendar.set(GregorianCalendar.YEAR, year);
		calendar.set(GregorianCalendar.MONTH, month - 1);
		calendar.set(GregorianCalendar.DAY_OF_MONTH, day);
		calendar.set(GregorianCalendar.HOUR_OF_DAY, hour);
		calendar.set(GregorianCalendar.MINUTE, minute);
		calendar.set(GregorianCalendar.SECOND, second);
		calendar.set(GregorianCalendar.MILLISECOND, 0);

		mDateValue = calendar.getTimeInMillis();

		mYear = year;
		mMonth = month;
		mDay = day;
		mHour = hour;
		mMinute = minute;
		mSecond = second;

		mDayOfWeek = calendar.get(GregorianCalendar.DAY_OF_WEEK);
	}


	/**
	 * Return a DateTime whose time is rounded to the nearest minute.
	 * Returns our own object if no rounding is necessary.
	 */
	public DateTime roundToMinute()
	{
		if (mSecond == 0)
			return this;

		GregorianCalendar calendar = getCalendar();

		calendar.setTimeInMillis(mDateValue);

		calendar.set(GregorianCalendar.SECOND, 0);
		calendar.set(GregorianCalendar.MILLISECOND, 0);

		if (mSecond >= 30)
			calendar.add(GregorianCalendar.MINUTE, 1);

		return new DateTime(calendar.getTimeInMillis());
	}


	/**
	 * Return a DateTime whose time is rounded down to the specified
	 * multiple of an hour boundary (e.g., the multiple of 1 rounds
	 * down to the nearest hour).  Only small multiples are handled.
	 * Returns our own object if no rounding is necessary.
	 * @return
	 */
	public DateTime roundDownToHourMultiple(int multiple)
	{
		if ((mMinute == 0) && (mSecond == 0) &&
			((mHour % multiple) == 0))
		{
			return this;
		}

		GregorianCalendar calendar = getCalendar();

		calendar.setTimeInMillis(mDateValue);

		calendar.set(GregorianCalendar.MINUTE, 0);
		calendar.set(GregorianCalendar.SECOND, 0);
		calendar.set(GregorianCalendar.MILLISECOND, 0);
		calendar.add(GregorianCalendar.HOUR, -(mHour % multiple));

		return new DateTime(calendar.getTimeInMillis());
	}


	/**
	 * Return a DateTime whose time is rounded up to the specified
	 * multiple of an hour boundary (e.g., the multiple of 1 rounds
	 * down to the nearest hour).  Only small multiples are handled.
	 * @return
	 */
	public DateTime roundUpToHourMultiple(int multiple)
	{
		if ((mMinute == 0) && (mSecond == 0) &&
			((mHour % multiple) == 0))
		{
			return this;
		}

		GregorianCalendar calendar = getCalendar();

		calendar.setTimeInMillis(mDateValue);

		calendar.set(GregorianCalendar.MINUTE, 0);
		calendar.set(GregorianCalendar.SECOND, 0);
		calendar.set(GregorianCalendar.MILLISECOND, 0);
		calendar.add(GregorianCalendar.HOUR_OF_DAY, multiple - (mHour % multiple));

		return new DateTime(calendar.getTimeInMillis());
	}


	/**
	 * Return a DateTime whose time is set to "noon".
	 * That is, the hour is set to 12 and the fractional hour is cleared.
	 */
	public DateTime roundToNoon()
	{
		if ((mMinute == 0) && (mSecond == 0) && (mHour == 12))
			return this;

		GregorianCalendar calendar = getCalendar();

		calendar.setTimeInMillis(mDateValue);

		calendar.set(GregorianCalendar.MINUTE, 0);
		calendar.set(GregorianCalendar.SECOND, 0);
		calendar.set(GregorianCalendar.MILLISECOND, 0);
		calendar.set(GregorianCalendar.HOUR_OF_DAY, 12);

		return new DateTime(calendar.getTimeInMillis());
	}


	/**
	 * Return a new DateTime object whose broken down values
	 * represent the local time for the specified time zone.
	 * Such an object should only be used for extracting its
	 * individual date values for display, not for its raw time.
	 */
	public DateTime getLocalTime(TimeZone timezone)
	{
		int offset = timezone.getOffset(mDateValue);

		return new DateTime(mDateValue + offset);
	}


	/**
	 * Return a new DateTime object whose time is adjusted by
	 * adding the specified number of milliseconds.
	 */
	public DateTime getAdjustedTime(long offset)
	{
		if (offset == 0)
			return this;

		return new DateTime(mDateValue + offset);
	}


	/**
	 * Return our Gregorian Calendar instance whose values are cleared
	 * and which has its timezone set to GMT (UTC).  The returned calendar
	 * is reused on each call and so it should not be kept by the caller.
	 * @return
	 */
	private GregorianCalendar getCalendar()
	{
		mCalendar.clear();
		mCalendar.setTimeZone(mGMTimeZone);

		return mCalendar;
	}


	public int getYear()
	{
		return mYear;
	}


	public int getMonth()
	{
		return mMonth;
	}


	public int getDay()
	{
		return mDay;
	}


	public int getHour()
	{
		return mHour;
	}


	public int getMinute()
	{
		return mMinute;
	}


	public int getSecond()
	{
		return mSecond;
	}


	public long getDateValue()
	{
		return mDateValue;
	}


	/**
	 * Return the day of the week as a string.
	 * This is the full name of the day, but extracting the first three
	 * letters will give the abbreviation.
	 * @return
	 */
	public String getDayOfWeekText()
	{
		switch (mDayOfWeek)
		{
			case GregorianCalendar.SUNDAY:
				return "Sunday";

			case GregorianCalendar.MONDAY:
				return "Monday";

			case GregorianCalendar.TUESDAY:
				return "Tuesday";

			case GregorianCalendar.WEDNESDAY:
				return "Wednesday";

			case GregorianCalendar.THURSDAY:
				return "Thursday";

			case GregorianCalendar.FRIDAY:
				return "Friday";

			case GregorianCalendar.SATURDAY:
				return "Saturday";

			default:
				return "Unknown";
		}
	}


	/**
	 * Return the number of minutes into the day.
	 */
	public int getMinutesIntoDay()
	{
		return mHour * 60 + mMinute;
	}


	/**
	 * Return the number of minutes from this time to another one.
	 * This might not be exactly right, but should work.
	 * @param other
	 * @return
	 */
	public int getMinutesTo(DateTime other)
	{
		double diff = ((double) (other.mDateValue - mDateValue)) / (1000 * 60);

		return (int) Math.round(diff);
	}


	/**
	 * Return the time offset from this time to another one.
	 */
	public long getOffsetTo(DateTime other)
	{
		return other.mDateValue - mDateValue;
	}


	/**
	 * Return whether this observation is in the same "day" as another one.
	 * This is correct for GMT time values, but for other time zones the time
	 * needs to be adjusted to the local time before this is meaningful.
	 * @param other
	 * @return
	 */
	public boolean isSameDay(DateTime other)
	{
		return ((mYear == other.mYear) && (mMonth == other.mMonth) &&
			(mDay == other.mDay));
	}


	/**
	 * Return whether this observation is in the same "month" as another one.
	 * This is correct for GMT time values, but for other time zones the time
	 * needs to be adjusted to the local time before this is meaningful.
	 * @param other
	 * @return
	 */
	public boolean isSameMonth(DateTime other)
	{
		return ((mYear == other.mYear) && (mMonth == other.mMonth));
	}


	@Override
	public int compareTo(DateTime other)
	{
		if (mDateValue < other.mDateValue)
			return -1;

		if (mDateValue > other.mDateValue)
			return 1;

		return 0;
	}


	@Override
	public boolean equals(Object obj)
	{
		if (! (obj instanceof DateTime))
			return false;

		DateTime other = (DateTime) obj;

		return mDateValue == other.mDateValue;
	}


	@Override
	public int hashCode()
	{
		return (int) (mDateValue / (1000 * 60));
	}


	/**
	 * Return the date as an easy to read string in the format:
	 *	yyyy/mm/dd hh:mm:ss
	 */
	public String getFormattedString()
	{
		return String.format("%04d/%02d/%02d %02d:%02d:%02d",
			mYear, mMonth, mDay, mHour, mMinute, mSecond);
	}


	/**
	 * Return the date value as a numeric string in the format:
	 * 	yyyymmddhhmmss
	 * or
	 * 	yyyymmddhhmm
	 */
	@Override
	public String toString()
	{
		String format = "%04d%02d%02d%02d%02d%02d";

		if (mSecond == 0)
			format = "%04d%02d%02d%02d%02d";

		return String.format(format, mYear, mMonth, mDay,
			mHour, mMinute, mSecond);
	}
}
