/**
 * Copyright (c) 2013 David I. Bell
 * Permission is granted to use, distribute, or modify this source,
 * provided that this copyright notice remains intact.
 */
package weather.db;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;

import weather.Observation;
import weather.ObservationList;
import weather.DateTime;
import weather.Utils;


/**
 * Class to write observation data to the database.
 * The database is a simple one composed of text files where each line
 * is one observation.  The text files are collected under directories
 * having the site names.  The site names are contained within one
 * directory defined by the configuration file.  Each text file only has
 * data for one (UTC) month and has a name like yyyy-mm.dat.
 * So a full file name is like:
 *	databasepath/siteName/yyyy-mm.dat
 */
public class ObservationWriter
{
	private File mDatabaseDirectory;
	private File mFile;
	private FileWriter mFileWriter;


	/**
	 * Write out the observations in the specified observation list.
	 * If the file is not to contain duplicates then the observation list
	 * must be pruned to remove observations already known to be
	 * in the file.
	 * @param observationList
	 * @throws IOException
	 */
	public void writeObservations(ObservationList observationList) throws IOException
	{
		if (observationList.isEmpty())
			return;

		//
		// Create the top level database directory if necessary.
		//
		mDatabaseDirectory = DBUtils.getDatabaseDirectory();

		DBUtils.makeDirectory(mDatabaseDirectory);

		//
		// Iterate over the sorted observation list while it is not empty.
		// We process the observations in chunks having the same site
		// and month, and so the non-matching observations are
		// saved in a temporary list and processed in the next loop.
		//
		observationList.sort();

		while (!observationList.isEmpty())
		{
			//
			// Get information about the first observation.
			// This determines the set of observations to be saved
			// into its database file with the others like it.
			//
			Observation firstObservation = observationList.getFirstObservation();
			DateTime firstUTCTime = firstObservation.getUTCTime();

			openFile(firstObservation);

			//
			// Process the observation list, either saving each observation
			// into the currently opened database file or adding it to the
			// skipped list for processing in the next loop.
			//
			ObservationList skippedObservationList =
				new ObservationList(observationList.size());

			for (Observation observation : observationList)
			{
				if (observation.isSameSite(firstObservation) &&
					observation.getUTCTime().isSameMonth(firstUTCTime))
				{
					appendToFile(observation);
				}
				else
				{
					skippedObservationList.add(observation);
				}
			}

			closeFile();

			//
			// Copy the skipped observations for the next loop.
			//
			observationList = skippedObservationList;
		}
	}


	/**
	 * Open the correct database file for appending based on the
	 * information from the specified observation.  The observation
	 * data itself is not written to the file here, but few lines of
	 * comments describing the data fields are written.
	 * @param observation
	 */
	private void openFile(Observation observation) throws IOException
	{
		//
		// Close any previous file just in case.
		//
		closeFile();

		//
		// Get the site name and data file name from this observation.
		//
		String siteName = DBUtils.getSiteDirectoryName(observation);
		String fileName = DBUtils.getDataFileName(observation);

		//
		// Create the site name directory if required and then
		// open the data file for appending.
		//
		File siteDirectory = new File(mDatabaseDirectory, siteName);

		DBUtils.makeDirectory(siteDirectory);

		mFile = new File(siteDirectory, fileName);

		boolean isCreated = !mFile.exists();

		mFileWriter = new FileWriter(mFile, true);

		//
		// If the file is created then write comment lines giving the format.
		//
		if (isCreated)
		{
			mFileWriter.append("# siteName, latitude, longitude, UTCTime, localTime, shortTimeString,\n");
			mFileWriter.append("# airTemp, apparentTemp, deltaTemp, dewPoint, relativeHumidity,\n");
			mFileWriter.append("# windSpeedKmh, gustKmh, windDir, pressureQNH, pressureMSL, rain\n");
		}
	}


	/**
	 * Close the file being written to if necessary, catching exceptions.
	 */
	private void closeFile()
	{
		Utils.safeClose(mFileWriter);

		mFileWriter = null;
		mFile = null;
	}


	/**
	 * Append the data in the specified observation to the currently
	 * opened file.  The site and date in the observation should match
	 * that of the file being written to.
	 * @param observation
	 */
	private void appendToFile(Observation observation) throws IOException
	{
		//
		// Get the string values of the observations.
		// Empty strings are set where the values are null.
		//
		String siteName = getStringTextValue(observation.getSiteName());
		String latitude = getDoubleTextValue(observation.getLatitude());
		String longitude = getDoubleTextValue(observation.getLongitude());
		String UTCTime = getDateTextValue(observation.getUTCTime());
		String localTime = getDateTextValue(observation.getLocalTime());
		String shortTimeString = getStringTextValue(observation.getShortTimeString());
		String airTemp = getDoubleTextValue(observation.getAirTemp());
		String apparentTemp = getDoubleTextValue(observation.getApparentTemp());
		String deltaTemp = getDoubleTextValue(observation.getDeltaTemp());
		String dewPoint = getDoubleTextValue(observation.getDewPoint());
		String relativeHumidity = getDoubleTextValue(observation.getRelativeHumidity());
		String windSpeedKmh = getDoubleTextValue(observation.getWindSpeedKmh());
		String gustKmh = getDoubleTextValue(observation.getGustKmh());
		String windDir = getStringTextValue(observation.getWindDir());
		String pressureQNH = getDoubleTextValue(observation.getPressureQNH());
		String pressureMSL = getDoubleTextValue(observation.getPressureMSL());
		String rain = getDoubleTextValue(observation.getRain());

		//
		// Format the values separated by commas, and append the line.
		//
		String line = String.format(
			"%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s\n",
			siteName, latitude, longitude, UTCTime, localTime, shortTimeString,
			airTemp, apparentTemp, deltaTemp, dewPoint, relativeHumidity,
			windSpeedKmh, gustKmh, windDir, pressureQNH, pressureMSL, rain);

		mFileWriter.append(line);
	}


	private String getStringTextValue(String value)
	{
		if (value == null)
			return "";

		value = value.replace(',', '.');

		return value;
	}


	/**
	 * Return the text value representing a double value while
	 * handling negative zero and removing any ".0" suffix.
	 */
	private String getDoubleTextValue(Double value)
	{
		if (value == null)
			return "";

		if (value == 0)
			return "0";

		String text = String.valueOf(value);

		if (text.endsWith(".0"))
			text = text.substring(0, text.length() - 2);

		if (text.isEmpty())
			text = "0";

		return text;
	}


	private String getDateTextValue(DateTime value)
	{
		if (value == null)
			return "";

		return value.toString();
	}
}
