/**
 * Copyright (c) 2013 David I. Bell
 * Permission is granted to use, distribute, or modify this source,
 * provided that this copyright notice remains intact.
 */
package weather.gui;

import java.awt.Container;
import java.awt.Frame;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.HashMap;
import java.util.Set;
import java.util.TreeMap;

import javax.swing.BorderFactory;
import javax.swing.BoxLayout;
import javax.swing.ButtonGroup;
import javax.swing.JDialog;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JRadioButton;
import javax.swing.border.TitledBorder;

import weather.Configuration;
import weather.Controller;
import weather.Utils;


/**
 * A dialog which displays a menu of actions.
 */
public class ActionsDialog extends JDialog implements ActionListener
{
	private static final long serialVersionUID = 1L;

	private static final String ACTION_INITIAL_DAYS = "Initial Days";
	private static final String ACTION_LATEST_REPORT = "Latest Report";
	private static final String ACTION_ONE_DAY = "One Day";
	private static final String ACTION_TWO_DAYS = "Two Days";
	private static final String ACTION_THREE_DAYS = "Three Days";
	private static final String ACTION_FOUR_DAYS = "Four Days";
	private static final String ACTION_FIVE_DAYS = "Five Days";
	private static final String ACTION_SEVEN_DAYS = "Seven Days";
	private static final String ACTION_FOURTEEN_DAYS = "Fourteen Days";
	private static final String ACTION_THIRTY_DAYS = "Thirty Days";
	private static final String ACTION_ALL_OBSERVATIONS = "All Observations";

	private static final String ACTION_UNIFORM_SCALING = "Uniform Scaling";

	private static final String ACTION_SITE_PREFIX = "Site ";

	private Controller mController;
	private Configuration mConfig;
	private JPanel mTimeIntervalPanel;
	private JPanel mSitePanel;
	private JPanel mOptionsPanel;

	private ButtonGroup mTimeIntervalButtonGroup;
	private ButtonGroup mSiteButtonGroup;

	private HashMap<String, JRadioButton> mSiteButtonMap;


	/**
	 * Constructor for the dialog which specifies the owning frame
	 * and the application Controller.  The dialog is not yet shown.
	 * @param frame
	 * @param controller
	 */
	public ActionsDialog(Frame frame, Controller controller)
	{
		super(frame, "Actions");

		mController = controller;

		mController.setActionsDialog(this);

		mConfig = Configuration.getInstance();

		createTimeIntervalPanel();
		createSitePanel();
		createOptionsPanel();

		Container content = getContentPane();

		content.setLayout(new BoxLayout(content, BoxLayout.X_AXIS));

		content.add(mTimeIntervalPanel);
		content.add(mSitePanel);
		content.add(mOptionsPanel);

		setLocation(500, 300);
		pack();
	}


	/**
	 * Create a panel to allow the selection of the site to display.
	 */
	private void createSitePanel()
	{
		mSitePanel = new JPanel();

		mSitePanel.setLayout(new BoxLayout(mSitePanel, BoxLayout.Y_AXIS));

		mSitePanel.setBorder(BorderFactory.createTitledBorder(
			BorderFactory.createEtchedBorder(),
			"Weather Sites", TitledBorder.CENTER, TitledBorder.TOP));

		//
		// Get the list of sites and sort them by the site names.
		//
		Set<String> siteKeySet = mConfig.getSiteMapKeys();

		TreeMap<String, String> sortedSiteMap = new TreeMap<String, String>();

		for (String siteKey : siteKeySet)
		{
			String siteName = mConfig.getSiteName(siteKey);

			sortedSiteMap.put(siteName, siteKey);
		}

		//
		// Now use the sorted map to create the buttons in sorted order.
		//
		mSiteButtonMap = new HashMap<String, JRadioButton>();
		mSiteButtonGroup = new ButtonGroup();

		for (String siteName : sortedSiteMap.keySet())
		{
			String siteKey = sortedSiteMap.get(siteName);

			//
			// Add the mnenomic key to the button name if applicable.
			//
			if ((siteKey.length() == 1)  && Utils.isLetter(siteKey.charAt(0)))
				siteName += "  [" + siteKey + "]";

			createSiteButton(siteKey, siteName, ACTION_SITE_PREFIX + siteKey);
		}
	}


	private void createSiteButton(String siteKey, String siteName, String actionCommand)
	{
		JRadioButton radioButton = new JRadioButton(siteName);

		radioButton.setActionCommand(actionCommand);
		radioButton.addActionListener(this);

		mSiteButtonGroup.add(radioButton);

		mSitePanel.add(radioButton);

		mSiteButtonMap.put(siteKey, radioButton);
	}


	/**
	 * Create a panel to allow the selection of the time interval to display.
	 */
	private void createTimeIntervalPanel()
	{
		mTimeIntervalPanel = new JPanel();

		mTimeIntervalPanel.setLayout(new BoxLayout(mTimeIntervalPanel, BoxLayout.Y_AXIS));

		mTimeIntervalPanel.setBorder(BorderFactory.createTitledBorder(
			BorderFactory.createEtchedBorder(),
			"Time Intervals", TitledBorder.CENTER, TitledBorder.TOP));

		mTimeIntervalButtonGroup = new ButtonGroup();

		int initialDays = mConfig.getTimeIntervalDays();

		String dayText = ((initialDays == 1) ? " day" : " days");
		String initialText = ((initialDays == 0) ? "latest report" : String.valueOf(initialDays) + dayText);

		createTimeIntervalButton("Latest report", ACTION_LATEST_REPORT);
		createTimeIntervalButton("All observations", ACTION_ALL_OBSERVATIONS);
		createTimeIntervalButton("Initial (" + initialText + ")", ACTION_INITIAL_DAYS);
		mTimeIntervalPanel.add(new JLabel(" "));
		createTimeIntervalButton("One day", ACTION_ONE_DAY);
		createTimeIntervalButton("Two days", ACTION_TWO_DAYS);
		createTimeIntervalButton("Three days", ACTION_THREE_DAYS);
		createTimeIntervalButton("Four days", ACTION_FOUR_DAYS);
		createTimeIntervalButton("Five days", ACTION_FIVE_DAYS);
		createTimeIntervalButton("Seven days", ACTION_SEVEN_DAYS);
		createTimeIntervalButton("Fourteen days", ACTION_FOURTEEN_DAYS);
		createTimeIntervalButton("Thirty days", ACTION_THIRTY_DAYS);
	}


	private void createTimeIntervalButton(String name, String actionCommand)
	{
		JRadioButton radioButton = new JRadioButton(name);

		radioButton.setActionCommand(actionCommand);
		radioButton.addActionListener(this);

		if (actionCommand.equals(ACTION_INITIAL_DAYS))
			radioButton.setSelected(true);

		mTimeIntervalButtonGroup.add(radioButton);

		mTimeIntervalPanel.add(radioButton);
	}


	/**
	 * Create an options panel to enable or disable particular features.
	 */
	private void createOptionsPanel()
	{
		mOptionsPanel = new JPanel();

		mOptionsPanel.setLayout(new BoxLayout(mOptionsPanel, BoxLayout.Y_AXIS));

		mOptionsPanel.setBorder(BorderFactory.createTitledBorder(
			BorderFactory.createEtchedBorder(),
			"Options", TitledBorder.CENTER, TitledBorder.TOP));

		JRadioButton radioButton = new JRadioButton("Use uniform scaling");

		radioButton.setActionCommand(ACTION_UNIFORM_SCALING);
		radioButton.addActionListener(this);

		radioButton.setSelected(mConfig.isUniformScaling());

		mOptionsPanel.add(radioButton);
	}


	/**
	 * Set the site button for the specified site key to be selected.
	 * @param siteKey
	 */
	public void setSelectedSite(String siteKey)
	{
		JRadioButton button = mSiteButtonMap.get(siteKey);

		if (button != null)
			button.setSelected(true);
	}


	/**
	 * Action handler for the components in the dialog.
	 */
	@Override
	public void actionPerformed(ActionEvent e)
	{
		String command = e.getActionCommand();
		Object source = e.getSource();

		//
		// If this action is for of the site buttons then
		// set the site to be displayed.
		//
		if (command.startsWith(ACTION_SITE_PREFIX))
		{
			String siteKey = command.substring(ACTION_SITE_PREFIX.length());

			mController.setReport(siteKey);

			setVisible(false);

			return;
		}

		//
		// Check for the special all observations action.
		//
		if (command.equals(ACTION_ALL_OBSERVATIONS))
		{
			mController.setTimeInterval(Long.MAX_VALUE);

			setVisible(false);

			return;
		}

		//
		// Look for options.
		//
		if (command.equals(ACTION_UNIFORM_SCALING))
		{
			JRadioButton button = (JRadioButton) source;

			mController.setUniformScaling(button.isSelected());

			setVisible(false);

			return;
		}

		//
		// Look for other specific actions.
		//
		long days = -1;

		if (command.equals(ACTION_INITIAL_DAYS))
			days = mConfig.getTimeIntervalDays();

		if (command.equals(ACTION_LATEST_REPORT))
			days = 0;

		if (command.equals(ACTION_ONE_DAY))
			days = 1;

		if (command.equals(ACTION_TWO_DAYS))
			days = 2;

		if (command.equals(ACTION_THREE_DAYS))
			days = 3;

		if (command.equals(ACTION_FOUR_DAYS))
			days = 4;

		if (command.equals(ACTION_FIVE_DAYS))
			days = 5;

		if (command.equals(ACTION_SEVEN_DAYS))
			days = 7;

		if (command.equals(ACTION_FOURTEEN_DAYS))
			days = 14;

		if (command.equals(ACTION_THIRTY_DAYS))
			days = 30;

		//
		// The days value is being changed.
		// Tell the controller the new interval and update the plot.
		//
		if (days >= 0)
		{
			mController.setTimeInterval(days * 24 * 60 * 60 * 1000);

			setVisible(false);
		}
	}
}


