/**
 * Copyright (c) 2013 David I. Bell
 * Permission is granted to use, distribute, or modify this source,
 * provided that this copyright notice remains intact.
 */
package weather.gui;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.Stroke;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.geom.AffineTransform;
import java.util.Date;
import java.util.TimeZone;

import javax.swing.JFrame;
import javax.swing.JPanel;

import weather.Configuration;
import weather.Controller;
import weather.DateTime;
import weather.Observation;
import weather.ObservationList;
import weather.Report;
import weather.SunriseSunset;
import weather.Utils;
import weather.selector.Selector;


/**
 * A panel to plot observations.
 * This can show either one or two plots at the same time.
 * If two plots are shown and are compatible (i.e., they have the same units)
 * than the labels and grid lines are meaningful for both plots.  Otherwise the
 * second plot is shown without any labels or grid.
 */
public class PlottingPanel extends JPanel
{
	private static final long serialVersionUID = 1L;

	private int mHashedLineSize = 4;
	private int mHashedLineInterval = 10;
	private int mSideInset = 40;
	private int mTopInset = 10;
	private int mBottomInset = 50;

	private Controller mController;

	private Graphics2D mGraphics;
	private int mGridWidth;
	private int mGridHeight;

	private int mPointRadius;
	private int mPointShrinkageDays;
	private int mWindDirectionRadius;
	private Color mBackgroundColor;
	private Color mNightTimeBackgroundColor;
	private Color mDayTimeBackgroundColor;
	private Color mMajorHorizontalLineColor;
	private Color mMinorHorizontalLineColor;
	private Color mMajorVerticalLineColor;
	private Color mMinorVerticalLineColor;
	private Color mLabelColor;
	private Color mDayOfWeekColor;

	private float [] mDashes = { 1f, 1f, 1f };

	private BasicStroke mDashesStroke = new BasicStroke(1,
		BasicStroke.CAP_BUTT, BasicStroke.JOIN_BEVEL, 1.0f, mDashes, 2f);

	private Stroke mDefaultStroke;
	private AffineTransform mDefaultTransform;

	//
	// The minimum and maximum UTC times for the plot.
	// This can be before or after any of the observations due to rounding.
	//
	private DateTime mMaximumTime;
	private DateTime mMinimumTime;

	//
	// The table of the sunrise and sunset times for the current observation list.
	// The lengths of these tables are the number of "noons" that the times use.
	//
	private long[] mSunriseTable;
	private long[] mSunsetTable;

	private Observation mSelectedObservation;
	private boolean mIsFirstObservationPlotted;
	private boolean mIsLastObservationPlotted;

	/**
	 * The report which this plot uses.
	 * The observation list which is plotted can contain more or less
	 * data than what is contained in the report.
	 */
	private Report mReport;

	/**
	 * The information about the first plot (which is always the primary one).
	 * This plot determines the type of value which is shown by the vertical
	 * labels and determines the horizontal grid lines for the major and
	 * minor intervals.
	 */
	private final PlotInfo mPlotInfo1 = new PlotInfo();

	/**
	 * The information about the second plot (which is optional).
	 * If the value types are compatible with the first plot then this
	 * plot shares the labels and grid lines with the first plot.
	 * If the value types are incompatible then the second plot will
	 * be shown but without any grid lines or labels of its own.
	 */
	private final PlotInfo mPlotInfo2 = new PlotInfo();

	private boolean mIsConsistent;
	private JFrame mOwnerFrame;

	private int mTimeMajorInterval;
	private int mTimeMinorInterval;

	private TimeZone mGMTTimezone = TimeZone.getTimeZone("GMT");


	/**
	 * Constructor which accepts the owning frame and our controller.
	 * @param ownerFrame
	 * @param controller
	 */
	public PlottingPanel(JFrame ownerFrame, Controller controller)
	{
		mOwnerFrame = ownerFrame;
		mController = controller;

		mController.setPlottingPanel(this);

		Configuration config = Configuration.getInstance();

		setPreferredSize(new Dimension(config.getPlotWidth(), config.getPlotHeight()));

		mPointRadius = config.getPointRadius();
		mPointShrinkageDays = config.getPointShrinkageDays();
		mWindDirectionRadius = config.getWindDirectionRadius();

		getColors();

		setBackground(mBackgroundColor);

		setEnabled(true);
		setFocusable(true);

		//
		// Add event handlers for the keyboard and mouse.
		//
		addKeyListener(new KeyAdapter()
			{
				@Override
				public void keyPressed(KeyEvent e)
				{
					handleKeyPressed(e);
				}

				@Override
				public void keyTyped(KeyEvent e)
				{
					handleKeyTyped(e);
				}
			});

		addMouseListener(new MouseAdapter()
			{
				@Override
				public void mouseClicked(MouseEvent e)
				{
					handleMouseClicked(e);
				}
			});
	}


	/**
	 * Set a report to to be shown in the panel along with the
	 * specified observation lists (which may or may not be the
	 * original ones from the report), and the intended time interval
	 * to be plotted and the number of hours to shift the plot by.
	 * @param report
	 */
	public void setReport(Report report,
		ObservationList fullObservationList,
		ObservationList observationList,
		ObservationList observationList2,
		DateTime minimumTime, DateTime maximumTime,
		int timeMajorInterval, int timeMinorInterval,
		boolean isFirstObservationPlotted,
		boolean isLastObservationPlotted)
	{
		mReport = report;
		mPlotInfo1.mFullObservationList = fullObservationList;
		mPlotInfo1.mObservationList = observationList;
		mPlotInfo2.mFullObservationList = fullObservationList;
		mPlotInfo2.mObservationList = observationList;
		mMinimumTime = minimumTime;
		mMaximumTime = maximumTime;
		mTimeMajorInterval = timeMajorInterval;
		mTimeMinorInterval = timeMinorInterval;
		mIsFirstObservationPlotted = isFirstObservationPlotted;
		mIsLastObservationPlotted = isLastObservationPlotted;

		mIsConsistent = observationList.isConsistent();

		//
		// Calculate the sunrise and sunset times for the
		// time interval.
		//
		calculateSunriseSunsetTimes();

		update();
	}


	/**
	 * Update the view when no new report or observation list has
	 * been set but something has changed to affect the title or plot.
	 * For examples, the type of observation being plotted or whether
	 * uniform scaling is applied.
	 */
	public void update()
	{
		setTitle();
		repaint();
	}


	/**
	 * Set the selectors for the item types to be plotted.
	 * The first selector must be non-null, but the second selector
	 * can be null if a second plot is not wanted.
	 */
	public void setSelectors(Selector selector1, Selector selector2)
	{
		mPlotInfo1.mSelector = selector1;
		mPlotInfo2.mSelector = selector2;

		update();
	}


	/**
	 * Set the title for the owning Frame to display the report info
	 * and the selector info.
	 */
	private void setTitle()
	{
		if ((mReport == null) || !mPlotInfo1.isValid())
			return;

		String title = mReport.getSiteName() + ": " +
			mReport.getRefreshMessage() + " - " +
			mPlotInfo1.mSelector.getName();

		//
		// Add the information about the second plot if any.
		//
		if (mPlotInfo2.isValid() && (mPlotInfo2.mSelector != mPlotInfo1.mSelector))
		{
			title += " + " + mPlotInfo2.mSelector.getName();

			if (mPlotInfo1.mSelector.getValueType() != mPlotInfo2.mSelector.getValueType())
				title += " (NO COORDINATES)";
		}

		mOwnerFrame.setTitle(title);
	}


	/**
	 * Draw this component.
	 */
	public void paintComponent(Graphics graphics)
	{
		mGraphics = (Graphics2D) graphics;

		mDefaultStroke = mGraphics.getStroke();
		mDefaultTransform = mGraphics.getTransform();

		mGraphics.setColor(mBackgroundColor);
		mGraphics.fillRect(0, 0, getWidth(), getHeight());

		if ((mReport == null) || !mPlotInfo1.isValid())
			return;

		//
		// Make sure that the first plot has at least one value to be plotted.
		// Check the full observation list instead of the current one so that
		// gaps in the observations can be skipped over.
		//
		if (!mPlotInfo1.mSelector.hasValues(mPlotInfo1.mFullObservationList))
			return;

		//
		// Something exists to be drawn.
		// Calculate the size of the plotting grid.
		//
		mSideInset = mPlotInfo1.mSelector.getSideInset();

		mGridWidth = getWidth() - mSideInset * 2;
		mGridHeight = getHeight() - mTopInset - mBottomInset;

		if ((mGridWidth <= 0) || (mGridHeight <= 0))
			return;

		//
		// Calculate the minimum and maximum values to be plotted
		// and the major and minor intervals for that range.
		//
		calculateValuesRange();

		//
		// Now draw the plot components in the correct order.
		//
		drawNightTimeBackground();
		drawDayTimeBackground();
		drawMinorVerticalLines();
		drawMajorVerticalLines();
		drawHashedEdges();
		drawMinorHorizontalLines(mPlotInfo1);
		drawMajorHorizontalLines(mPlotInfo1);
		drawHorizontalLabels(mPlotInfo1);
		drawSelectedValues(mPlotInfo2);
		drawAnnotations(mPlotInfo2);
		drawSelectedValues(mPlotInfo1);
		drawAnnotations(mPlotInfo1);

		mGraphics.setStroke(mDefaultStroke);
	}


	/**
	 * Calculate the minimum and maximum values to be plotted
	 * and the major and minor intervals for that range.
	 * This handles the case of either one plot or two plots.
	 * In the case of two plots, they can be related or unrelated.
	 */
	private void calculateValuesRange()
	{
		//
		// Calculate the ranges for the two plots assuming they are unrelated.
		//
		calculateValuesRange(mPlotInfo1);
		calculateValuesRange(mPlotInfo2);

		//
		// See whether there is a second plot and whether its values
		// are compatible with the first plot.  If not then do no more.
		//
		if (!mPlotInfo1.isValid() || !mPlotInfo2.isValid())
			return;

		if (!mPlotInfo2.mSelector.hasValues(mPlotInfo2.mObservationList))
			return;

		if (mPlotInfo1.mSelector.getValueType() !=
			mPlotInfo2.mSelector.getValueType())
		{
			return;
		}

		//
		// The second plot is compatible, so calculate new minimum
		// and maximum values to cover both of their ranges so that they
		// can both fit and use the same grid and labels.
		//
		double minimumValue =
			Math.min(mPlotInfo1.mMinimumValue, mPlotInfo2.mMinimumValue);

		double maximumValue =
			Math.max(mPlotInfo1.mMaximumValue, mPlotInfo2.mMaximumValue);

		//
		// Get the range for both plots and determine the major and minor
		// intervals for the range.  Only the first selector is used since it is the
		// primary one, and any second selector should be compatible.
		//
		double range = maximumValue - minimumValue;

		double majorInterval = mPlotInfo1.mSelector.getMajorInterval(range);
		double minorInterval = mPlotInfo1.mSelector.getMinorInterval(range);

		//
		// Round the minimum and maximum values to the major intervals.
		//
		maximumValue = Utils.roundUpToExactMultiple(maximumValue, majorInterval);
		minimumValue = Utils.roundDownToExactMultiple(minimumValue, majorInterval);

		//
		// Set the values into both plot information instances so that
		// they share the same values.
		//
		mPlotInfo1.mMinimumValue = minimumValue;
		mPlotInfo1.mMaximumValue = maximumValue;
		mPlotInfo1.mMajorInterval = majorInterval;
		mPlotInfo1.mMinorInterval = minorInterval;

		mPlotInfo2.mMinimumValue = minimumValue;
		mPlotInfo2.mMaximumValue = maximumValue;
		mPlotInfo2.mMajorInterval = majorInterval;
		mPlotInfo2.mMinorInterval = minorInterval;
	}


	/**
	 * Calculate the minimum and maximum values to be plotted
	 * and the major and minor intervals for that range.
	 * The range can be calculated either for the observations currently
	 * being shown in the plot, or for all observations that could be
	 * plotted by shifting.
	 */
	private void calculateValuesRange(PlotInfo plotInfo)
	{
		if (!plotInfo.isValid())
			return;

		//
		// Get the observation list to use for calculating the minimum
		// and maximum values depending on whether uniform
		// scaling is enabled.
		//
		ObservationList observationList = plotInfo.mObservationList;

		if (mController.isUniformScaling())
			observationList = plotInfo.mFullObservationList;

		//
		// Get the minimum and maximum values for the plot.
		//
		double minimumValue = plotInfo.mSelector.getMinimumValue(observationList);
		double maximumValue = plotInfo.mSelector.getMaximumValue(observationList);

		//
		// Get the range for both plots and determine the major and minor
		// intervals for the range.
		//
		double range = maximumValue - minimumValue;

		plotInfo.mMajorInterval = plotInfo.mSelector.getMajorInterval(range);
		plotInfo.mMinorInterval = plotInfo.mSelector.getMinorInterval(range);

		//
		// Round the minimum and maximum values to the major intervals.
		//
		plotInfo.mMaximumValue = Utils.roundUpToExactMultiple(
			maximumValue, plotInfo.mMajorInterval);

		plotInfo.mMinimumValue = Utils.roundDownToExactMultiple(
			minimumValue, plotInfo.mMajorInterval);

		//
		// If the range is empty then make it at least one major interval high.
		//
		if (plotInfo.mMaximumValue <= plotInfo.mMinimumValue)
		{
			plotInfo.mMaximumValue =
				plotInfo.mMinimumValue + plotInfo.mMajorInterval;
		}
	}


	/**
	 * Draw the minor vertical lines for the grid.
	 * These lines are dashed.
	 */
	private void drawMinorVerticalLines()
	{
		mGraphics.setColor(mMinorVerticalLineColor);
		mGraphics.setStroke(mDashesStroke);

		double minutesInterval = mMinimumTime.getMinutesTo(mMaximumTime);

		double horizontalInterval = minutesInterval / mTimeMinorInterval;

		int horizontalDivisions = (int) Math.round(Math.abs(horizontalInterval));

		for (int i = 0; i <= horizontalDivisions; i++)
		{
			int xPos = mSideInset + (mGridWidth * i) / horizontalDivisions;

			mGraphics.drawLine(xPos, mTopInset, xPos, mTopInset + mGridHeight);
		}

		mGraphics.setStroke(mDefaultStroke);
	}


	/**
	 * Draw the major vertical lines for the grid.
	 * This also draws the labels for the date/time at the bottom.
	 */
	private void drawMajorVerticalLines()
	{
		double minutesInterval = mMinimumTime.getMinutesTo(mMaximumTime);
		double horizontalInterval = minutesInterval / mTimeMajorInterval;

		int horizontalDivisions = (int) Math.round(Math.abs(horizontalInterval));

		for (int i = 0; i <= horizontalDivisions; i++)
		{
			double range = mMaximumTime.getDateValue() - mMinimumTime.getDateValue();
			double value = mMinimumTime.getDateValue() + (range * i) / horizontalDivisions;

			DateTime GMTTime = new DateTime((long) value).roundToMinute();

			//
			// Draw a normal or a wide line.
			//
			boolean isWide = false;

			if ((GMTTime.getMinutesTo(mMaximumTime) % (60 * 24)) == 0)
				isWide = true;

			int xPos = mSideInset + (mGridWidth * i) / horizontalDivisions;

			mGraphics.setColor(mMajorVerticalLineColor);

			mGraphics.drawLine(xPos, mTopInset, xPos, mTopInset + mGridHeight);

			if (isWide)
			{
				mGraphics.drawLine(xPos - 1, mTopInset, xPos - 1, mTopInset + mGridHeight);
				mGraphics.drawLine(xPos + 1, mTopInset, xPos + 1, mTopInset + mGridHeight);
			}

			//
			// Draw the rotated label at the bottom showing the local time.
			// On day boundaries show the hour and day, otherwise just
			// show the hour.
			//
			DateTime localTime = getLocalTime(GMTTime);

			String label;

			if (localTime.getHour() == 0)
				label = String.format("%02d %02d", localTime.getHour(), localTime.getDay());
			else
				label = String.format("%02d", localTime.getHour());

			mGraphics.setColor(mLabelColor);
			mGraphics.rotate(Math.PI / 2);
			mGraphics.drawString(label, mTopInset + mGridHeight + 8, -xPos + 4);
			mGraphics.setTransform(mDefaultTransform);

			//
			// Draw the day of the week underneath the noon-time label.
			// Use its own color so that it isn't confused with the other labels.
			//
			if ((localTime.getHour() == 12) && (localTime.getMinute() == 0) &&
				(mTimeMinorInterval < 24 * 60))
			{
				label = localTime.getDayOfWeekText().substring(0, 3);

				mGraphics.setColor(mDayOfWeekColor);
				mGraphics.drawString(label, xPos - 10, mTopInset + mGridHeight + mBottomInset - 8);
			}
		}
	}


	/**
	 * Draw the minor horizontal lines for the grid.
	 * These lines are dashed.
	 */
	private void drawMinorHorizontalLines(PlotInfo plotInfo)
	{
		mGraphics.setColor(mMinorHorizontalLineColor);
		mGraphics.setStroke(mDashesStroke);

		double range = plotInfo.mMaximumValue - plotInfo.mMinimumValue;
		double verticalInterval =range / plotInfo.mMinorInterval;

		int verticalDivisions = (int) Math.round(Math.abs(verticalInterval));

		for (int i = 0; i <= verticalDivisions; i++)
		{
			int yPos = mTopInset + (mGridHeight * i ) / verticalDivisions;

			mGraphics.drawLine(mSideInset, yPos, mSideInset + mGridWidth, yPos);
		}

		mGraphics.setStroke(mDefaultStroke);
	}


	/**
	 * Draw the major horizontal lines for the grid.
	 */
	private void drawMajorHorizontalLines(PlotInfo plotInfo)
	{
		double range = plotInfo.mMaximumValue - plotInfo.mMinimumValue;
		double verticalInterval = range / plotInfo.mMajorInterval;
		int verticalDivisions = (int) Math.round(Math.abs(verticalInterval));

		mGraphics.setColor(mMajorHorizontalLineColor);

		for (int i = 0; i <= verticalDivisions; i++)
		{
			int yPos = mTopInset + (mGridHeight * i ) / verticalDivisions;

			mGraphics.drawLine(mSideInset, yPos, mSideInset + mGridWidth, yPos);
		}
	}


	/**
	 * Draw the labels for the horizontal grid lines on the left and right sides.
	 * These can be shown for just the major intervals or for both the major
	 * and minor intervals depending on the selector's choice.
	 */
	private void drawHorizontalLabels(PlotInfo plotInfo)
	{
		double range = plotInfo.mMaximumValue - plotInfo.mMinimumValue;

		double plotInterval = (plotInfo.mSelector.isMinorIntervalLabeled(range) ?
			plotInfo.mMinorInterval : plotInfo.mMajorInterval);

		double verticalInterval = range / plotInterval;
		int verticalDivisions = (int) Math.round(Math.abs(verticalInterval));

		mGraphics.setColor(mLabelColor);

		for (int i = 0; i <= verticalDivisions; i++)
		{
			int yPos = mTopInset + (mGridHeight * i ) / verticalDivisions;

			//
			// Format and draw the labels at the left and right sides.
			// Suppress the printing of a negative zero.
			//
			double value = plotInfo.mMaximumValue - (range * i) / verticalDivisions;

			String label = String.format("%.1f", value);

			if (label.equals("-0.0"))
				label = "0.0";

			mGraphics.drawString(label, 4, yPos + 4);
			mGraphics.drawString(label, mSideInset + mGridWidth + 6, yPos + 6);
		}
	}


	/**
	 * If necessary draw hashed left or right edges for the grid to indicate
	 * that shifting may be done to see more of the plot in that direction.
	 * This is based on whether the plot shows the first and last observations.
	 */
	private void drawHashedEdges()
	{
		if (!mIsFirstObservationPlotted)
			drawHashedLine(mSideInset);

		if (!mIsLastObservationPlotted)
			drawHashedLine(mSideInset + mGridWidth);
	}


	/**
	 * Draw hashes on the line at the specified X coordinate running from the
	 * bottom to the the top of the grid.  The hashes are short diagonal lines
	 * crossing the vertical line.
	 * @param XPosition
	 */
	private void drawHashedLine(int x)
	{
		mGraphics.setColor(mMajorVerticalLineColor);

		for (int y = mHashedLineInterval / 2;
			y < mGridHeight - mHashedLineInterval / 2;
			y += mHashedLineInterval)
		{
			mGraphics.drawLine(x - mHashedLineSize,
				y + mTopInset + mHashedLineSize,
				x + mHashedLineSize,
				y + mTopInset - mHashedLineSize);
		}
	}


	/**
	 * Draw the night time background.
	 * This is done before drawing the daytime so we can just fill the
	 * whole grid with the night time color.
	 */
	private void drawNightTimeBackground()
	{
		mGraphics.setColor(mNightTimeBackgroundColor);

		mGraphics.fillRect(mSideInset, mTopInset, mGridWidth, mGridHeight);
	}


	/**
	 * Calculate the sunrise and sunset times for the current observation list.
	 * The results are saved into the mSunRise and mSunSet arrays.
	 * This is only done if all the observations are consistent (i.e, the same site).
	 */
	private void calculateSunriseSunsetTimes()
	{
		if (!mIsConsistent)
			return;

		Observation firstObservation = mPlotInfo1.mObservationList.getFirstObservation();
		Observation lastObservation = mPlotInfo1.mObservationList.getLastObservation();

		if ((firstObservation == null) || (lastObservation == null))
			return;

		//
		// Calculate the number of "noons" which covers the time
		// range of the observation list.
		//
		DateTime firstNoon = getNoonTime(firstObservation.getUTCTime());
		DateTime lastNoon = getNoonTime(lastObservation.getUTCTime());

		long firstNoonTime = firstNoon.getDateValue();
		long lastNoonTime = lastNoon.getDateValue();

		long noonDifference = lastNoonTime - firstNoonTime;

		int noonCount = ((int) Math.ceil(((double) noonDifference) / (24*60*60*1000))) + 1;

		mSunriseTable = new long[noonCount];
		mSunsetTable = new long[noonCount];

		//
		// Now calculate the sunrise and sunset times which surrounds
		// each of the noon times and save the results into the arrays.
		//
		SunriseSunset sunriseSunset = new SunriseSunset();

		double latitude = lastObservation.getLatitude();
		double longitude = lastObservation.getLongitude();
		long noonTime = firstNoonTime;

		for (int index = 0; index < noonCount; index++)
		{
			Date time = new Date(noonTime);

			Date sunrise = sunriseSunset.getSunrise(latitude, longitude, time, mGMTTimezone);
			Date sunset = sunriseSunset.getSunset(latitude, longitude, time, mGMTTimezone);

			//
			// Fix off by one day calculation mistakes.
			//
			if (sunrise.getTime() > time.getTime())
				sunrise = new Date(sunrise.getTime() - 24*60*60*1000);

			if (sunset.getTime() < time.getTime())
				sunset = new Date(sunset.getTime() + 24*60*60*1000);

			mSunriseTable[index] = sunrise.getTime();
			mSunsetTable[index] = sunset.getTime();

			noonTime += 24*60*60*1000;
		}
	}


	/**
	 * Color the day time in the background using a lighter color than the night time.
	 * The night time background has already been drawn.
	 * This is only done if all the observations are consistent (i.e, the same site).
	 * This uses the previously calculated sunrise and sunset times.
	 */
	private void drawDayTimeBackground()
	{
		if (!mIsConsistent)
			return;

		mGraphics.setColor(mDayTimeBackgroundColor);

		for (int index = 0; index < mSunriseTable.length; index++)
		{
			DateTime minTime = new DateTime(mSunriseTable[index]);
			DateTime maxTime = new DateTime(mSunsetTable[index]);

			if (minTime.compareTo(mMinimumTime) < 0)
				minTime = mMinimumTime;

			if (maxTime.compareTo(mMaximumTime) > 0)
				maxTime = mMaximumTime;

			if (minTime.compareTo(maxTime) >= 0)
				continue;

			int fromX = getXCoordinate(minTime);
			int toX = getXCoordinate(maxTime);

			mGraphics.fillRect(fromX, mTopInset, toX - fromX + 1, mGridHeight);
		}
	}


	/**
	 * Calculate the GMT time of the local noon from the specified GMT time.
	 * @param time
	 * @return
	 */
	private DateTime getNoonTime(DateTime GMTTime)
	{
		DateTime localTime = getLocalTime(GMTTime);

		long GMTToLocalOffset = GMTTime.getOffsetTo(localTime);

		localTime = localTime.roundToNoon();

		return localTime.getAdjustedTime(-GMTToLocalOffset);
	}


	/**
	 * Return the Local time corresponding to the specified GMT time value.
	 * This uses the time offset calculated from the closest Observation to
	 * the GMT time.  This is close to correct, except possibly for times near
	 * the exact time of changes between standard and daylight savings.
	 * (We do this instead of trying to use a TimeZone because the timezone
	 * names in the collected reports are unusable.)
	 * @param GMTTime
	 */
	private DateTime getLocalTime(DateTime GMTTime)
	{
		Observation closestObservation = mPlotInfo1.mFullObservationList.getClosestObservation(GMTTime);

		if (closestObservation == null)
			return GMTTime;

		long GMTToLocalOffset = closestObservation.getLocalTimeOffset();

		DateTime localTime = GMTTime.getAdjustedTime(GMTToLocalOffset);

		return localTime;
	}


	/**
	 * Draw the observation values for the specified plot.
	 * The data points are drawn as circles with lines connecting them
	 * if successive observations are considered connected.
	 */
	private void drawSelectedValues(PlotInfo plotInfo)
	{
		if (!plotInfo.isValid())
			return;

		Selector selector = plotInfo.mSelector;

		DateTime prevTime = null;
		double prevValue = 0;
		Observation prevObservation = null;

		for (Observation observation : plotInfo.mObservationList)
		{
			if (!selector.hasDoubleValue(observation))
				continue;

			DateTime time = observation.getUTCTime();
			double value = selector.getDoubleValue(observation);

			drawPoint(plotInfo, time, value, (observation == mSelectedObservation));

			if (selector.isJoined(prevObservation, observation))
				drawLine(plotInfo, prevTime, prevValue, time, value);

			prevTime = time;
			prevValue = value;
			prevObservation = observation;
		}
	}


	/**
	 * Draw a circle representing an observation point.
	 * The circle's radius is reduced to just 1 pixel if the number of days
	 * shown in the plot is large enough.  The circle can be drawn open
	 * or filled.
	 */
	private void drawPoint(PlotInfo plotInfo, DateTime time,
		double value, boolean isFilled)
	{
		int x = getXCoordinate(time);
		int y = getYCoordinate(plotInfo, value);

		int radius = mPointRadius;

		if ((mPointShrinkageDays > 0) &&
			(mMinimumTime.getMinutesTo(mMaximumTime) >= mPointShrinkageDays * 60 * 24))
		{
			radius = 1;
		}

		mGraphics.setColor(plotInfo.mPointColor);

		if (isFilled)
		{
			mGraphics.fillArc(x - radius, y - radius, radius * 2, radius * 2, 0, 360);
		}
		else
		{
			mGraphics.drawArc(x - radius, y - radius, radius * 2, radius * 2, 0, 360);
		}
	}


	private void drawLine(PlotInfo plotInfo, DateTime time1, double value1,
		DateTime time2, double value2)
	{
		int x1 = getXCoordinate(time1);
		int x2 = getXCoordinate(time2);
		int y1 = getYCoordinate(plotInfo, value1);
		int y2 = getYCoordinate(plotInfo, value2);

		mGraphics.setColor(plotInfo.mLineColor);
		mGraphics.drawLine(x1, y1, x2, y2);
	}


	/**
	 * Draw annotations for the observations.
	 * This is a special-case method.
	 * The only annotations are to show wind direction
	 * vectors for the wind speed plot.
	 */
	private void drawAnnotations(PlotInfo plotInfo)
	{
		if (!plotInfo.isValid())
			return;

		//
		// Annotations are only applied to the wind speed
		// and only if the wind direction radius is nonzero.
		//
		if (plotInfo.mSelector != Selector.WIND_SPEED_KMH_SELECTOR)
			return;

		if (mWindDirectionRadius <= 0)
			return;

		mGraphics.setColor(plotInfo.mAnnotationColor);

		for (Observation observation : plotInfo.mObservationList)
		{
			//
			// If there isn't any wind speed then ignore
			// this observation.
			//
			if (!plotInfo.mSelector.hasDoubleValue(observation))
				continue;

			//
			// Get the bearing of the wind direction.
			// If there isn't any then ignore this observation.
			//
			Double bearing = Selector.WIND_DIRECTION_SELECTOR.
				getBearing(observation);

			if (bearing == null)
				continue;

			//
			// Calculate the center of the wind speed point.
			//
			DateTime time = observation.getUTCTime();
			double value = plotInfo.mSelector.getDoubleValue(observation);

			int x = getXCoordinate(time);
			int y = getYCoordinate(plotInfo, value);

			//
			// Calculate the vector to draw based on the
			// bearing and draw a line from the center of
			// the circle outwards.
			//
			double angle = Math.toRadians(90.0 - bearing);

			int rx = (int) Math.round(Math.cos(angle) * mWindDirectionRadius);
			int ry = (int) Math.round(Math.sin(angle) * mWindDirectionRadius);

			mGraphics.drawLine(x, y, x + rx, y - ry);
		}
	}


	/**
	 * Get the X coordinate for drawing in the panel based on the given date value.
	 * This is usually a value within the displayed grid.
	 * @param time
	 * @return
	 */
	private int getXCoordinate(DateTime time)
	{
		int range = mMinimumTime.getMinutesTo(mMaximumTime);
		int used = mMinimumTime.getMinutesTo(time);

		double scaledValue = ((double) used) / range;

		int coordinate = ((int) Math.round(scaledValue * mGridWidth)) + mSideInset;

		return coordinate;
	}


	/**
	 * Get the Y coordinate for drawing in the panel based on the given value.
	 * This is usually a value within the displayed grid.
	 */
	private int getYCoordinate(PlotInfo plotInfo, double value)
	{
		double range = plotInfo.mMaximumValue - plotInfo.mMinimumValue;
		double used = value - plotInfo.mMinimumValue;

		double scaledValue = used / range;

		int coordinate = mTopInset + mGridHeight - ((int) Math.round(scaledValue * mGridHeight));

		return coordinate;
	}


	/**
	 * Set the Color values as defined in the configuration file.
	 * Default colors are used in case of missing or illegal values.
	 */
	private void getColors()
	{
		Configuration config = Configuration.getInstance();

		//
		// The order of obtaining these matters a bit since some
		// default colors use the previously obtained colors.
		//
		mBackgroundColor = GUIUtils.getColorFromName(
			config.getBackgroundColor(),
			Color.blue.darker().darker().darker().darker());

		mDayTimeBackgroundColor = GUIUtils.getColorFromName(
			config.getDayTimeBackgroundColor(),
			Color.blue.darker());

		mNightTimeBackgroundColor = GUIUtils.getColorFromName(
			config.getNightTimeBackgroundColor(),
			mDayTimeBackgroundColor.darker());

		mMajorHorizontalLineColor = GUIUtils.getColorFromName(
			config.getMajorHorizontalLineColor(),
			Color.green.darker());

		mMinorHorizontalLineColor = GUIUtils.getColorFromName(
			config.getMinorHorizontalLineColor(),
			mMajorHorizontalLineColor.darker());

		mMajorVerticalLineColor = GUIUtils.getColorFromName(
			config.getMajorVerticalLineColor(),
			Color.green.darker());

		mMinorVerticalLineColor = GUIUtils.getColorFromName(
			config.getMinorVerticalLineColor(),
			mMajorVerticalLineColor.darker());

		mPlotInfo1.mPointColor = GUIUtils.getColorFromName(
			config.getPointColor(),
			Color.yellow);

		mPlotInfo1.mLineColor = GUIUtils.getColorFromName(
			config.getLineColor(),
			mPlotInfo1.mPointColor);

		mPlotInfo1.mAnnotationColor = GUIUtils.getColorFromName(
			config.getAnnotationColor(),
			Color.red);

		mPlotInfo2.mPointColor = GUIUtils.getColorFromName(
			config.getPointColor2(),
			mPlotInfo1.mPointColor.darker().darker().darker());

		mPlotInfo2.mLineColor = GUIUtils.getColorFromName(
			config.getLineColor2(),
			mPlotInfo1.mLineColor.darker().darker().darker());

		mPlotInfo2.mAnnotationColor = GUIUtils.getColorFromName(
			config.getAnnotationColor2(),
			mPlotInfo1.mAnnotationColor.darker().darker());

		mLabelColor = GUIUtils.getColorFromName(
			config.getLabelColor(),
			Color.yellow);

		mDayOfWeekColor = GUIUtils.getColorFromName(
			config.getDayOfWeekColor(),
			Color.yellow.darker().darker().darker());
	}


	/**
	 * Method called on a key typed event.
	 * @param event
	 */
	protected void handleKeyTyped(KeyEvent event)
	{
		char ch = event.getKeyChar();
		int modifiers = event.getModifiersEx();

		boolean isControlDown =
			((modifiers & KeyEvent.CTRL_DOWN_MASK) != 0);

		switch (ch)
		{
			case 'q':
				mController.exit();
				return;

			case 'u':
				mController.updateReport();
				return;

			default:
				break;
		}

		if ((ch >= '0') && (ch <= '9'))
		{
			if (isControlDown)
				mController.addSelector(ch - '0');
			else
				mController.setSelector(ch - '0');

			return;
		}

		//
		// If the character is a letter then set the report
		// using a site key of that letter.
		//
		if (Utils.isLetter(ch))
		{
			mController.setReport(String.valueOf(ch));
		}

		System.out.println("Keychar " + ch + " modifiers " + modifiers);
	}


	/**
	 * Method called on a key pressed event.
	 * This is used to shift the view left or right by the major interval
	 * (or by days if shift/control is used).
	 */
	protected void handleKeyPressed(KeyEvent event)
	{
		int keyCode = event.getKeyCode();
		int modifiers = event.getModifiersEx();

		int shiftHours = mTimeMajorInterval / 60;

		if ((modifiers & (KeyEvent.CTRL_DOWN_MASK | KeyEvent.SHIFT_DOWN_MASK)) != 0)
			shiftHours = 24;

		switch (keyCode)
		{
			case KeyEvent.VK_LEFT:
				mController.shiftViewLeft(shiftHours);
				break;

			case KeyEvent.VK_RIGHT:
				mController.shiftViewLeft(-shiftHours);
				break;
		}
	}


	/**
	 * Method called on a mouse clicked event.
	 * @param event
	 */
	protected void handleMouseClicked(MouseEvent event)
	{
		int button = event.getButton();
		Point point = event.getPoint();

		if (button == 1)
		{
			handleLeftClick(point.x, point.y);

			return;
		}

		if (button == 3)
		{
			handleRightClick(point.x, point.y);

			return;
		}
	}


	/**
	 * Handle a left click event on at the specified coordinates.
	 * This is used to bring up a details dialog box for the selected
	 * observation or to shift the view.
	 */
	private void handleLeftClick(int targetX, int targetY)
	{
		//
		// Shift left if the labels at the left are clicked.
		//
		if (targetX < mSideInset)
		{
			int hours = mMinimumTime.getMinutesTo(mMaximumTime) / 60;
			mController.shiftViewLeft(hours);

			return;
		}

		//
		// Shift right if the labels at the right are clicked.
		//
		if (targetX > mSideInset + mGridWidth)
		{
			int hours = mMinimumTime.getMinutesTo(mMaximumTime) / 60;
			mController.shiftViewLeft(-(hours));

			return;
		}

		//
		// The click is within the grid so search for the nearest observation.
		// If one is found then tell the controller to select it.
		//
		Observation observation = findObservation(mPlotInfo1, targetX, targetY);

		if (observation == null)
			observation = findObservation(mPlotInfo2, targetX, targetY);

		if (observation != null)
			mController.selectObservation(observation);
	}


	/**
	 * Find the best observation which is close to the specified coordinates.
	 * Returns null if the coordinates are not close to any observation.
	 */
	private Observation findObservation(PlotInfo plotInfo, int targetX, int targetY)
	{
		if (!plotInfo.isValid())
			return null;

		Observation bestObservation = null;
		int bestDistance = 0;

		for (Observation observation : plotInfo.mObservationList)
		{
			if (!plotInfo.mSelector.hasDoubleValue(observation))
				continue;

			int x = getXCoordinate(observation.getUTCTime());
			int y = getYCoordinate(plotInfo, plotInfo.mSelector.getDoubleValue(observation));

			int xDiff = x - targetX;
			int yDiff = y - targetY;

			int distance = xDiff * xDiff + yDiff * yDiff;

			if (distance > mPointRadius * mPointRadius)
				continue;

			if ((bestObservation == null) || (distance < bestDistance))
			{
				bestObservation = observation;
				bestDistance = distance;
			}
		}

		return bestObservation;
	}


	/**
	 * Handle a right click event at the specified coordinates.
	 * This is used to bring up an actions dialog.
	 */
	private void handleRightClick(int targetX, int targetY)
	{
		mController.showActionsDialog();
	}


	/**
	 * Called by the controller to set or clear the selected observation.
	 * For us, this simply means to fill the selected point, or to clear
	 * the selected point.  A redraw is needed to show the change.
	 * @param observation
	 */
	public void setObservation(Observation observation)
	{
		if (mSelectedObservation != observation)
		{
			mSelectedObservation = observation;

			update();
		}
	}


	/**
	 * A convenience class to hold all of the related information for a single plot.
	 * This information is only used for the plotting of the points and lines for
	 * the observation values, and keeps no data about the labels or time. 
	 */
	private static class PlotInfo
	{
		public ObservationList mFullObservationList;
		public ObservationList mObservationList;
		public Selector mSelector;
		public Color mLineColor;
		public Color mPointColor;
		public Color mAnnotationColor;
		public double mMaximumValue;
		public double mMinimumValue;
		public double mMajorInterval;
		public double mMinorInterval;


		/**
		 * Return whether or not this plot is valid for use.
		 */
		public boolean isValid()
		{
			return ((mObservationList != null) && (mSelector != null));
		}
	}
}
